#include <CoreFoundation/CoreFoundation.h>

#include <iostream>
#include <string>
#include <vector>
#include <chrono>
#include <thread>
#include <functional>
#include <dlfcn.h>

#include "OPN_Interface.hpp"
#include "OpticonDeviceEvent.hpp"
#include "OpticonDataEvent.hpp"


/**
 *  \brief Demo application that demonstrates how to use various API calls of the OPN library.
 *
 */
class OPNDEMO : public EventHandler<OpticonDeviceEvent>, public EventHandler<OpticonDataEvent>
{
public:
    
    /**
     *  \brief Constructor
     */
    OPNDEMO(OPN_Interface* opn){
        //save the externally created pointer
        this->opn = opn;
    }
    
    /**
     *  \brief Destructor
     */
    ~OPNDEMO(){
        //remove the listeners
        opn->removeDataEventListener(dataReg);
        opn->removeDeviceEventListener(deviceReg);
    }
    
    /**
     *  This method is called when a device event (insert and remove) occurs.
     *
     * @param e     Opticon device event
     */
    virtual void onEvent(OpticonDeviceEvent & e) override {
        
        if(e.getEventKind() == kEventKindInsert) {
            std::cout << e.getMessage() << " " << e.getDeviceId() << std::endl;
            //add device
            deviceConnected(e.getDeviceId());
        }
        
        else if(e.getEventKind() == kEventKindRemove) {
            std::cout << e.getMessage() << std::endl;
            //remove device
            deviceDisconnected(e.getDeviceId());
        }
    }
    
    /**
     * \brief This is called when new data has been read
     *
     * @param e     The data event
     */
    virtual void onEvent(OpticonDataEvent & e) override {
        std::cout <<  e.getMessage() << std::endl;
        printBarcode(e.getData(), e.getHandler());
    }
    
    /**
     *  \brief Initialize the library and register for data and device events
     */
    void initialize(){
        //initialize the device
        opn->Initialize();
        //register the listeners so we can get all data and device events
        dataReg = opn->addDataEventListener(*this);
        deviceReg = opn->addDeviceEventListener(*this);
    }
    
    /**
     *  \brief Clean up
     */
    void cleanup(){
        
    }
    
    /**
     *  \brief  Print all barcodes
     */
    void printBarcode(std::vector<Barcode> data, int handler){
        std::cout << "Reading data from USB ID = " << handler << std::endl;
        
        for(auto it = data.begin(); it!= data.end(); it++)
        {
            Barcode barcode = *it;
            //print the barcode, timestamp and symbology
            std::cout << barcode.getData() << " " << barcode.getTimestamp() << " " << barcode.getSymbology() << std::endl;
        }
        std::cout << std::endl;
    }
    
    /**
     *  \brief Check if data is available. Note this does not check if there is new data but what was read last time (stale data).
     To check for the latest data call getAllBarcodes.
     */
    void checkForData(int handler){
        int res;
        bool retVal;
        std::vector<Barcode> myList;
        
        res = opn->isDataAvailable(handler, retVal);
        
        if(retVal && res == STATUS_OK){
            //get the data
            opn->getAllBarcodes(handler, myList);
            //print the data
            printBarcode(myList, handler);
        }
    }
    
    /**
     *  Exercise different commands...
     */
    void deviceConnected(int handler){
        //save the device handler for future reference
        deviceHandlers.push_back(handler);
        
        //interrogate the device
        interrogate(handler);
        
        //TEST time/date
        TEST_Timestamp(handler);
        
        //change param setting
        setParameter(handler);
        
        //get param
        getParameter(handler);
        
        //set default settings
        setDefaultSetting(handler);
        
        //manual check without handler
        ManualCheckForData();
        
    }
    
    /**
     *  \brief Interrogate the device
     */
    void interrogate(int handler){
        int res = opn->interrogate(handler);
        
        if(res == STATUS_OK){
            std::string version, libVer, devId;
            UInt32 sysStat;
            opn->getSoftwareVersion(handler, version);
            opn->getSystemStatus(handler, sysStat);
            opn->getLibraryVersion(libVer);
            opn->getDeviceID(handler, devId);
            
            std::cout << "Software version = " << version << " System status = " << sysStat << " Library version = " << libVer << " Device ID = " << devId << std::endl;
        }
        else
            std::cout << "ERROR: could not interrogate the device. error number = " << res << std::endl;
        
    }
    
    /**
     *  \brief Set a parameter
     */
    void setParameter(int handler){
        int res = opn->setParam(handler, PARAM_BUZZER_VOLUME, BUZZER_SOFT);
        
        if(res == STATUS_OK){
            std::cout << "Buzzer volume disabled successfully! " << std::endl;
        }
        else
            std::cout << "ERROR: Could not set the parameter setting. error number = " << res << std::endl;
    }
    
    /**
     *  \brief Read a parameter
     */
    void getParameter(int handler){
        int paramValue; //holds parameter setting return value
        
        //get the buzzer volume setting
        int res = opn->getParam(handler, PARAM_BUZZER_VOLUME, paramValue);
        
        if(res == STATUS_OK){
            std::cout << "Buzzer volume is " << paramValue << std::endl;
        }
        else
            std::cout << "ERROR: Could not read param value. error number = " << res << std::endl;
        
    }
    
    
    void TEST_Timestamp(int handler){
        //set the time first
        //Set custom time
        struct tm future;
        future.tm_sec = 30;      /* seconds after the minute [0-60] */
        future.tm_min = 30;      /* minutes after the hour [0-59] */
        future.tm_hour = 9;      /* hours since midnight [0-23] */
        future.tm_mday = 26;     /* day of the month [1-31] */
        future.tm_mon = 6;       /* months since January [0-11] */
        future.tm_year = 150;    /* years since 1900 */ /********set to future time 2050 ;) ***********/
        
        
        //set the time
        int res = opn->setTime(handler, future);
        
        if(res == STATUS_OK){
            //print the new date
            std::cout << "Setting Device Timestamp to: " << future.tm_mon+1 << "/" << future.tm_mday << "/" << future.tm_year+1900 << " " <<
            future.tm_hour << ":" <<  future.tm_min << ":" <<  future.tm_sec << std::endl;
        }
        else
            std::cout << "ERROR: could not set the time from device. error number = " << res << std::endl;
        
        
        //get the time and verify it matches what we set previously
        struct tm new_timestamp;
        
        //get the time
        res = opn->getTime(handler, new_timestamp);
        
        if(res == STATUS_OK){
            std::cout << "Device Timestamp is: " << new_timestamp.tm_mon << "/" << new_timestamp.tm_mday << "/" << new_timestamp.tm_year+1900 << " "
            <<  new_timestamp.tm_hour << ":" <<  new_timestamp.tm_min << ":" <<  new_timestamp.tm_sec << std::endl;
        }
        else
            std::cout << "ERROR: could not get the time from device. error number = " << res << std::endl;
        
        
        //set the timestamp to now
        time_t t = time(0);   // get time now
        struct tm now = *localtime( & t );
        
        //set the time
        res = opn->setTime(handler, now);
        
        if(res == STATUS_OK){
            //print the new date
            std::cout << "Setting Device Timestamp to now: " << now.tm_mon+1 << "/" << now.tm_mday << "/" << now.tm_year+1900 << " " <<
            now.tm_hour << ":" <<  now.tm_min << ":" <<  now.tm_sec << std::endl;
        }
        else
            std::cout << "ERROR: could not set the time from device. error number = " << res << std::endl;
        
        
    }
    
    /**
     *  \brief Get the time/date
     */
    void getTimestamp(int handler){
        
        
    }
    
    /**
     *  \brief Set time/date
     */
    void setTimestamp(int handler){
        
    }
    
    /**
     *  \brief Set device settings to default
     */
    void setDefaultSetting(int handler){
        int res = opn->setDefaults(handler);
        if(res == STATUS_OK){
            std::cout << "Set Defaults successful." << std::endl;
        }
        else
            std::cout << "ERROR: Set Defaults failed. error number = " << res << std::endl;
    }
    
    
    /**
     *  \brief Remove the device from the list
     */
    void deviceDisconnected(int handler){
        //iterate throught the device handler list and find the device handler id
        for (std::vector<int>::iterator it = deviceHandlers.begin(); it != deviceHandlers.end(); ++it)
        {
            //if found stop polling and delete the device handler from the list
            if(*it == handler){
                deviceHandlers.erase(it);
                break;
            }
        }
    }
    
    /**
     *  \brief Go through all handlers and check for data....
     */
    void ManualCheckForData(){
        int counter = 0;
        std::vector<Barcode> myList;
        bool retVal;
        
        while(1){
            //iterate throught the device handler list and check for data
            for (std::vector<int>::iterator it = deviceHandlers.begin(); it != deviceHandlers.end(); ++it)
            {
                opn->getAllBarcodes(*it, myList);
                opn->isDataAvailable(*it, retVal);
                if(retVal){
                    printBarcode(myList, *it);
                    //clear data
//                    opn->clearData(*it);
                }
                
            }
            //2 sec delay
            sleep(2);
        }
    }
    
    
    /* Private member variables */
private:
    //vector to store all devices detected
    //NOTE: user has to decide how they would like to keep track of all the OPN devices connected and communicate with each
    //device they are interested in. If they only expect one device to be connected at all times then
    //this could be a single long variable.
    std::vector<int> deviceHandlers;
    
    //pointer to the OPN library object
    OPN_Interface *opn;
    
    //pointer to a device registration handler. Can be used to register and delete an event listener.
    HandlerRegistration *deviceReg;
    
    //pointer to a data registration handler. Can be used to register and delete an event listener.
    HandlerRegistration *dataReg;
    
};


//================================================================================================
//	main
//================================================================================================
int main(int argc, const char *argv[])
{
    
    using std::cout;
    using std::cerr;
    
    std::cout << "Running OPN test application " << std::endl;
    // Open the library.
    void* lib_handle = dlopen("/Users/nimanourozi/Projects/OPN_Lib_Test/libOPN_Library.dylib", RTLD_LOCAL);
    if (!lib_handle) {
        exit(EXIT_FAILURE);
    }
    
    // Get the NewOPN_Interface function.
    OPN_Interface_creator* NewOPN_Interface = (OPN_Interface_creator*)dlsym(lib_handle, "NewOPN_Interface");
    if (!NewOPN_Interface) {
        exit(EXIT_FAILURE);
    }
    
    // Get the DeleteOPN_Interface function.
    OPN_Interface_disposer* DeleteOPN_Interface = (OPN_Interface_disposer*)dlsym(lib_handle, "DeleteOPN_Interface");
    if (!DeleteOPN_Interface) {
        exit(EXIT_FAILURE);
    }
    
    //get the new instance
    OPN_Interface* lib = NewOPN_Interface();
    
    //pass the newly created object to demo app
    OPNDEMO demo(lib);
    
    //initialize
    demo.initialize();
    
    /***************************************
    ****** prevent main from exiting *******
    ***************************************/
    while(1){
        sleep(1);
    }
    
    
    // Destroy lib object.
    DeleteOPN_Interface(lib);
    
    // Close the library.
    if (dlclose(lib_handle) != 0) {
        exit(EXIT_FAILURE);
    }
    
    return 0;
}




/******************************************************************************
 ************ SEPARATE EVENT HANDLER CLASS ************************************
 ******************************************************************************/
///**
// * \brief Simple example of an event handler class
// *
// * This snippet shows how to implement a device listener
// */
//class OpnDeviceListener : public EventHandler<OpticonDeviceEvent>
//{
//public:
//    OpnDeviceListener(OPNDEMO *app):app(app){}
//
//    virtual ~OpnDeviceListener(){}
//
//    /**
//     *
//     *
//     * @param e     An Opticon device event
//     */
//    virtual void onEvent(OpticonDeviceEvent & e) override {
//
//        // Ignore the event if it's already been canceled
//        //        if (e.getCanceled()) {
//        //            return;
//        //        }
//
//        if(e.getEventKind() == kEventKindInsert) {
//            std::cout << "DEVICE EVENT OCCURED " << e.getMessage() << std::endl;
//            app->deviceConnected(e.getDeviceId());
//        }
//
//        else if(e.getEventKind() == kEventKindRemove) {
//            std::cout << "DEVICE EVENT OCCURED " << e.getMessage() << std::endl;
//            app->deviceDisconnected(e.getDeviceId());
//        }
//    }
//
//private:
//    OPNDEMO     *app;
//};
//
//
///**
// * \brief Simple example of an event handler class
// *
// * This snippet shows how to implement a data listener
// */
//class OpnDataListener : public EventHandler<OpticonDataEvent>
//{
//public:
//    OpnDataListener(OPNDEMO *app):app(app)
//    { }
//
//    virtual ~OpnDataListener() { }
//
//    /**
//     * \brief This event handler keeps the player inside a specific border area
//     *
//     * @param e The PlayerMoveEvent event
//     */
//    virtual void onEvent(OpticonDataEvent & e) override {
//
//        // Ignore the event if it's already been canceled
//        //        if (e.getCanceled()) {
//        //            return;
//        //        }
//        app->readData();
//        std::cout << "DATA EVENT OCCURED " << e.getMessage() << std::endl;
//    }
//
//private:
//    OPNDEMO     *app;
//};
